/*
 * Modem.java
 *
 * Copyright (C) 2011 John Douyere (VK2ETA) - for Android platforms
 * Partially based on Modem.java from Pskmail by Per Crusefalk and Rein Couperus
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.AndFlmsg;

import android.content.*;
import android.content.res.Resources;
import android.media.*;
import android.media.MediaRecorder.*;
import java.util.regex.*;




public class Modem {


	public static final int RXMODEMIDLE = 0;
	public static final int RXMODEMSTARTING = 1;
	public static final int RXMODEMRUNNING = 2;
	public static final int RXMODEMPAUSED = 3;
	public static final int RXMODEMSTOPPING = 4;

	public static int modemState = RXMODEMIDLE;
	
	public static boolean modemThreadOn = true;
	
	public static int NumberOfOverruns = 0;
	private static AudioRecord audiorecorder = null;
	private static boolean RxON = false;
	private static int bufferSize = 0;
	private final static float sampleRate = 8000.0f;

	//Must match the rsid.h declaration: #define RSID_FFT_SIZE		1024
	public static final int RSID_FFT_SIZE = 1024;
	//Must match the rsid.h declaration: #define RSID_SAMPLE_RATE 11025.0
	public final static double RSID_SAMPLE_RATE = 11025.0f;
 
	private static boolean FirstBracketReceived = false;
	public static String MonitorString = "";
	public static boolean WrapLF = false;
	private static String BlockString = "";
	
//	private static SampleRateConversion myResampler;

	public static double frequency = 1500.0; 
	public static double rsidFrequency = 0;
    static double squelch = 20.0;
	static double metric = 50; //midrange

    //Waterfall interaction
	public static double[] WaterfallAmpl = new double[RSID_FFT_SIZE];
    //Semaphore for waterfall (new array of amplitudes needed for display)
    public static boolean newAmplReady = false;

	public static boolean stopTX = false;
	
	//The following value MUST Match the MAXMODES in modem.h
	public static final int MAXMODES = 300; 
	//List of modems and modes returned from the C++ modems
	public static int[] modemCapListInt = new int[MAXMODES];
	public static String[] modemCapListString = new String[MAXMODES];
	public static int numModes = 0;
	//Custom list of modes as slected in the preferences (can 
	//  include all modes above if "Custom List" not selected or all modes manually selected)
	public static int[] customModeListInt = new int[MAXMODES];
	public static String[] customModeListString = new String[MAXMODES];
	public static int customNumModes = 0;
	
	//JD debug
	private static boolean CModemInitialized = false;

	
	//Declaration of native classes
    public native static String 	createCModem(int modemCode);
    public native static String 	initCModem(double frequency);
    public native static String 	rxCProcess(short[] buffer, int length);
    public native static int 		setSquelchLevel(double squelchLevel);
	public native static double 	getMetric();
	public native static double 	getCurrentFrequency();
	public native static int	 	getCurrentMode();
	public native static String 	RsidCModemReceive( float[] myfbuffer, int length);
    public native static String 	createRxRsidModem();
    private native static int[] 	getModemCapListInt();
    private native static String[] 	getModemCapListString();

    static {
    	//Load the C++ modems library
    	System.loadLibrary("AndFlmsg_Head_CPP");
    }

    
    public static void updateWaterfall(double[] aFFTAmpl ) {
		System.arraycopy(aFFTAmpl, 0, WaterfallAmpl, 0, RSID_FFT_SIZE);
		Modem.newAmplReady = true;
		AndFlmsg.mHandler.post(AndFlmsg.updatewaterfall);
    }

	public static void ModemInit() {

		//Prepare RSID Receive
		//		myRxRSID =  new RxRSID();
		createRxRsidModem();
		//get modem list (int and string description)
		modemCapListInt = getModemCapListInt();
		modemCapListString = getModemCapListString();
		//Now find the end of modem list to know how many different modems codes we have
		Modem.numModes = MAXMODES; //Just in case
		for (int i = 0; i < MAXMODES; i++){
			if (modemCapListInt[i] == -1) {
				Modem.numModes = i;
				//Exit loop
				i = MAXMODES;
			}
		}
		//Sort by mode code to re-group modes of the same modem (as they are in two arrayrs in rsid_def.cxx)
		boolean swapped = true;
		int tmp;
		String tmpS;
		while (swapped) {
			swapped = false;
			for (int i = 0; i < numModes - 1; i++) {
				if (modemCapListInt[i] > modemCapListInt[i + 1]) {
					tmp = modemCapListInt[i];
					tmpS = modemCapListString[i];
					modemCapListInt[i] = modemCapListInt[i + 1];
					modemCapListString[i] = modemCapListString[i + 1];
					modemCapListInt[i + 1] = tmp;
					modemCapListString[i + 1] = tmpS;
					swapped = true;
				}
			}
		}
		//Do we want to use a custom list of modes?
		boolean useModeList = config.getPreferenceB("USEMODELIST", false);
		if (useModeList) {
			customNumModes = 0;
			boolean thatmode;
			for (int i = 0; i < numModes; i++) {
				thatmode = config.getPreferenceB("USE"+ modemCapListString[i], false);
				if (thatmode) {
					customModeListInt[customNumModes] = modemCapListInt[i];
					customModeListString[customNumModes++] = modemCapListString[i];
				}
			}
		} else {
			customModeListInt = modemCapListInt;
			customModeListString = modemCapListString;
			customNumModes = numModes;
		}
		//Check that we have at least one mode (if not, take the first in the capability list)
		if (customNumModes < 1) {
			customNumModes = 1;
			customModeListInt[0] = modemCapListInt[0];
			customModeListString[0] = modemCapListString[0];
			//Store that in the preferences to be consistent
		    //SharedPreferences.Editor editor = AndFlmsg.mysp.edit();
		    //editor.putBoolean("USE"+ modemCapListString[1], true);
		    // Commit the edits!
		    //editor.commit();
		}
		//JD Change to C++ resampler instead of quadratic resampler Java code
		//Initialize Re-sampling to 11025Hz for RSID, THOR and MFSK modems
		//		myResampler = new SampleRateConversion(11025.0 / 8000.0);
		SampleRateConversion.SampleRateConversionInit((float) (11025.0 / 8000.0));

	}


	//Returns the modem code given a modem name (String)
	public static int getMode (String mstring) {
		Resources res = AndFlmsg.myContext.getResources();
		int[] modeValues = res.getIntArray(R.array.modesValues);
		String[] modeString = res.getStringArray(R.array.modesStrings);
		int j = 0;
		for (int i = 0; i < modeValues.length; i++) {
			if (mstring == modeString[i]) j = i;
		}
		return modeValues[j];
	}

	//Returns the modem index in the array of modes available given a modem code
	public static int getModeIndex (int mcode) {
		int j = -1;
		for (int i = 0; i < customNumModes; i++) {
			if (mcode == customModeListInt[i]) j = i;
		}
		//In case we didn't find it, return the first mode in the list to avoid segment fault
		if (j == -1) {
			j = 0;
		}
		return j;
	}

	
	//Return the new mode code or the same if it hits the end of list either way
	public static int getModeUpDown(int currentMode, int increment) {
		//Find position of current mode
		int j = getModeIndex(currentMode);
		j += increment;
		//Circular list
		if (j < 0) j = customNumModes - 1;
		if (j >= customNumModes) j = 0;
		return customModeListInt[j];
	}

	
	private static void soundInInit() {
		bufferSize = (int) sampleRate; // 1 second of Audio max
		if (bufferSize < AudioRecord.getMinBufferSize((int) sampleRate, AudioFormat.CHANNEL_IN_MONO , AudioFormat.ENCODING_PCM_16BIT)) { 
			// Check to make sure buffer size is not smaller than the smallest allowed one 
			bufferSize = AudioRecord.getMinBufferSize((int) sampleRate, AudioFormat.CHANNEL_IN_MONO , AudioFormat.ENCODING_PCM_16BIT);
		}
		int ii = 20; //number of 1/4 seconds wait
		while (--ii > 0) {
			if (AndFlmsg.toBluetooth) {
				//Bluetooth hack (use voice call)
				audiorecorder = new AudioRecord(AudioSource.MIC, 8000, android.media.AudioFormat.CHANNEL_IN_MONO, 
						android.media.AudioFormat.ENCODING_PCM_16BIT, bufferSize);
			} else {
				audiorecorder = new AudioRecord(AudioSource.MIC, 8000, android.media.AudioFormat.CHANNEL_IN_MONO, 
						android.media.AudioFormat.ENCODING_PCM_16BIT, bufferSize);
			}
			if (audiorecorder.getState() == AudioRecord.STATE_INITIALIZED) {
				ii = 0;//ok done
			} else {
				if (ii < 16) { //Only if have to wait more than 1 seconds
					loggingclass.writelog("Waiting for Audio MIC availability..." , null, true);
				}
				try {
					Thread.sleep(250);//1/4 second
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
			}
		}
		if (audiorecorder.getState() != AudioRecord.STATE_INITIALIZED) {
			//JD add exception catch here
			System.out.println("Can't open audio MIC");
			loggingclass.writelog("Can't open Audio MIC " , null, true);
		}
	}



	public static void startmodem() {
		modemThreadOn = true;

		new Thread(new Runnable() {
			public void run() {


				while (modemThreadOn) {

					String rsidReturnedString;
					String modemReturnedString;
					modemState = RXMODEMSTARTING;
					double startproctime = 0;
					double endproctime = 0;
					int numSamples8K = 0;
					soundInInit();
					NumberOfOverruns = 0;
					audiorecorder.startRecording();
					RxON = true;
					Processor.restartRxModem.drainPermits();
					//Since the callback is not working, implement a while loop.
					short[] so8K = new short[bufferSize];
					int size12Kbuf =  (int) ((bufferSize + 1) * 11025.0 / 8000.0);
					//JD changed to float to match rsid.cxx code					
					//double[] so12K = new double[size12Kbuf];
					float[] so12K = new float[size12Kbuf];
					//Initialise modem

					
					String modemCreateResult = createCModem(Processor.RxModem);
					if (modemCreateResult.contains("ERROR")) {
						CModemInitialized = false;
					} else {
						CModemInitialized = true;
						//Initialize RX side of modem
						String frequencySTR = config.getPreference("AFREQUENCY","1000");
						double centerfreq = Integer.parseInt(frequencySTR);
						initCModem(centerfreq);
					}
					while (RxON) {
						endproctime = System.currentTimeMillis();
						double buffertime = (double) numSamples8K / 8000.0 * 1000.0; //in milliseconds
						if (numSamples8K > 0) Processor.cpuload = (int)(((double)(endproctime - startproctime)) / buffertime * 100);
						if (Processor.cpuload > 100) Processor.cpuload = 100;
						AndFlmsg.mHandler.post(AndFlmsg.updatecpuload);
						//JD try faster mode changes by having a smaller buffer to process						numSamples8K = audiorecorder.read(so8K, 0, 8000/4); //process only part of the buffer to avoid lumpy processing
						numSamples8K = audiorecorder.read(so8K, 0, 8000/8); //process only part of the buffer to avoid lumpy processing
						if (numSamples8K > 0) {
							modemState = RXMODEMRUNNING;
							startproctime = System.currentTimeMillis();
							//Process only if Rx is ON, otherwise discard (we have already decided to TX)
							if (RxON) {
								//Re-sample to 11025Hz for RSID, THOR and MFSK modems
								int numSamples12K = SampleRateConversion.Process(so8K, numSamples8K, so12K, size12Kbuf);
								//RxRSID always ON
								rsidReturnedString = RsidCModemReceive( so12K, numSamples12K);
								if (rsidReturnedString.startsWith("\nRSID:")) {
									//We have a new modem and/or centre frequency
									Modem.MonitorString += rsidReturnedString;
									//Update the RSID waterfall frequency too
									frequency = getCurrentFrequency();
									Processor.RxModem = getCurrentMode();
									AndFlmsg.mHandler.post(AndFlmsg.updatetitle);
								}
								//Sets the latest squelch level for the modem to use
								setSquelchLevel(squelch);
								//Then the RX modem
								modemReturnedString = rxCProcess(so8K, numSamples8K);
								//Retreive latest signal quality for display
						        metric = misc.decayavg(metric, getMetric(), 20);
						        //metric = getMetric();
						        AndFlmsg.mHandler.post(AndFlmsg.updatesignalquality);
						        //Now process all the characters received
						        for (int i = 0; i < modemReturnedString.length(); i++) {
						        	processRxChar(modemReturnedString.charAt(i));
						        }
							}
							//Post to monitor (Modem) window after each buffer processing
							//Add TX frame too if present
							if (Modem.MonitorString.length() > 0 || Processor.TXmonitor.length() > 0) {
								Processor.monitor += Modem.MonitorString + Processor.TXmonitor;
								Processor.TXmonitor = "";
								Modem.MonitorString = "";
								AndFlmsg.mHandler.post(AndFlmsg.addtomodem);
							}
						}
					}//while (RxON)
					//We dropped here on pause flag
					if (audiorecorder != null) {
						//Avoid some crashes on wrong state
						if (audiorecorder.getState() == AudioRecord.STATE_INITIALIZED) {
							if (audiorecorder.getRecordingState() == AudioRecord.RECORDSTATE_RECORDING) {
								audiorecorder.stop();
							}
							audiorecorder.release();
						}
					}
					modemState = RXMODEMPAUSED;
					//Marker for end of thread (Stop modem thread flag)
					if (!modemThreadOn) {
						modemState = RXMODEMIDLE;
						return;
					}
					//Now waits for a restart (or having this thread killed)
					Processor.restartRxModem.acquireUninterruptibly(1);
					//Make sure we don's have spare permits
					Processor.restartRxModem.drainPermits();
				}//while (modemThreadOn)
				//We dropped here on thread stop request
				modemState = RXMODEMIDLE;				
			} //run
		}).start(); //	new Thread(new Runnable() {

	}


	public static void stopRxModem() {
		modemThreadOn = false;
		RxON = false;
	}

	public static void pauseRxModem() {
		RxON = false;
	}

	public static void unPauseRxModem() {
    	Processor.restartRxModem.release(1);
	}

	static void changemode(int newMode) {
		//Stop the modem receiving side to prevent using the wrong values
		pauseRxModem();
		//Restart modem reception
		unPauseRxModem();

	}

	static void setFrequency(double rxfreq) {
		frequency = rxfreq;
	}
	
	static void reset() {
		String frequencySTR = config.getPreference("AFREQUENCY","1000");
		frequency = Integer.parseInt(frequencySTR);
		if (frequency < 500) frequency = 500;
		if (frequency > 2500) frequency = 2500;
		squelch = AndFlmsg.mysp.getFloat("SQUELCHVALUE", (float) 20.0);
	}
	
    /**
     * Send a mode command to the modem
     */
    public static void send_rsid_command(String s){

    	//VK2ETA: do nothing for now but keep refernce in case we decide to make 
    	//   RSID receive conditional in the future

    	//        String rsidstart="<cmd><rsid>";
    	//        String rsidend="</rsid></cmd>";
    	//        if (!s.equals("")){
    	//                Main.SendCommand += (rsidstart+s+rsidend);
    	//        }       
    }    
    public static void send_txrsid_command(String s){
    	if (s.equals("ON")) {
    		Processor.TXID = true; //let the modem sent it when required
    	} else {
    		Processor.TXID = false; 
    	}

    }    
	
    /**
     * @param squelchdiff the delta to add to squelch
     */
    public static void AddtoSquelch(double squelchdiff) {
        squelch += (squelch > 10) ? squelchdiff : squelchdiff / 2;
        if (squelch < 0) squelch =0;
        if (squelch >100) squelch =100;
		//store value into preferences
	      SharedPreferences.Editor editor = AndFlmsg.mysp.edit();
	      editor.putFloat("SQUELCHVALUE", (float) squelch);
	      // Commit the edits!
	      editor.commit();
    }

	public static void processRxChar (char inChar) {
	
		if (inChar > 127) {
			// todo: unicode encoding
			inChar = 0;
		}
		
		switch (inChar) {
			case 0:
				break; // do nothing
			case '[':
				if (Processor.ReceivingForm ||
					FirstBracketReceived) {
					BlockString += inChar;
				} else {			
					if (!FirstBracketReceived) {
						FirstBracketReceived = true;
						BlockString = "[";
						//Processor.PostToTerminal("FirstBracket\n");
					} 
				}
				//	            Main.DCD = 0;
				break;
			case ']':
				if (Processor.ReceivingForm) {
					BlockString += inChar;
					//Did we receive a new start of block before the end 
					//of the previous block? If so, restart reception
					if (BlockString.endsWith("[WRAP:beg]")) {
						BlockString = "[WRAP:beg]";
						Processor.PostToTerminal("\n\nReceiving New Message, discarding any partially received data\n");
					} else if (BlockString.endsWith("[WRAP:lf]")) {
						WrapLF = true;
					} else if (BlockString.endsWith("[WRAP:crlf]")) {
							WrapLF = false;
					} else if (BlockString.endsWith("[WRAP:end]")) {
						//We have a complete block, process it
						//Processor.PostToTerminal("End\n");
						Processor.processWrapBlock(BlockString);
						Processor.CrcString = "";
						Processor.FileNameString = "";
						Processor.ReceivingForm = false;
						FirstBracketReceived = false;
					} else {
						//Check for CRC section
						Pattern psc = Pattern.compile("WRAP:chksum ([0-9A-F]{4})");
						Matcher msc = psc.matcher(BlockString);
						if (msc.find()) {
							Processor.CrcString = msc.group(1);
							//Processor.PostToTerminal("Checksum <" + Processor.CrcString + ">\n");
						}
						//Check for file name section
						psc = Pattern.compile("\\[WRAP:fn (.+\\..+)\\]");
						msc = psc.matcher(BlockString);
						if (msc.find()) {
							if (BlockString.endsWith(msc.group(0))) {
								Processor.FileNameString = msc.group(1);
								Processor.PostToTerminal("\nReceiving File " + Processor.FileNameString);
								//Processor.PostToTerminal("\n" + msc.group(0) + "\n");
							}
						}
					}
				} else {
					BlockString += inChar;
					if (BlockString.endsWith("[WRAP:beg]")) {
						FirstBracketReceived = true;
						BlockString = "[WRAP:beg]";
						Processor.ReceivingForm = true;
						Processor.PostToTerminal("\n\nReceiving New Message");
					}

				}
				break;
			case 10:
				MonitorString += "\n";
				if (Processor.ReceivingForm || 
					FirstBracketReceived) {
					BlockString += inChar;
				}
				break;
			case 13: //ignore CR
				break;
			default:
				if (Processor.ReceivingForm || 
					FirstBracketReceived) {
					BlockString += inChar;
				}
				break;
		}	// end switch
		//Reset if header not found within
		//  1000 charaters of first bracket
		if (FirstBracketReceived && 
			!Processor.ReceivingForm &&
			BlockString.length() > 1000) {
			BlockString = BlockString.substring(900);
			if (!BlockString.contains("]"))  FirstBracketReceived = false;
		}
		//Reset if end marker not found within
		// 30,000 charaters of start
		//to-do: if found crc but not wrap end, process anyway
		if (Processor.ReceivingForm &&
			BlockString.length() > 30000)
		{
			Processor.CrcString = "";
			Processor.FileNameString = "";
			Processor.ReceivingForm = false;
			FirstBracketReceived = false;	
		}
		if (inChar > 31)
		{
			MonitorString += inChar;
		}
	}


	//In a separate thread so that the UI thread is not blocked during TX
	public static void sendln(String Sendline)
	{
		Runnable TxRun = new TxThread(Sendline);
		Sendline = "";
		new Thread(TxRun).start();
	}

	public static class TxThread implements Runnable {
		private String TxSendline = "";

		public TxThread(String Sendline) {
			TxSendline = Sendline;
		}

		public void run() {
			// see if tx active and DCD is off
			//JD Debug - FIX this: review DCD logic        if (Sendline.length() > 0 & !Processor.TXActive & Processor.DCD == 0) {
			if (TxSendline.length() > 0 & !Processor.TXActive) {

				if (Processor.DCDthrow == 0){
					try {
						Processor.TXActive = true;
						AndFlmsg.mHandler.post(AndFlmsg.updatetitle);   
						//Stop the modem receiving side
						pauseRxModem();
						//TX delay here from config popup (allows on-the-fly changes)
						int TxDelay = 0;
				    	String TxDelayString = config.getPreference("TXDELAY","0");
				    	TxDelay = Integer.parseInt(TxDelayString);

						if (TxDelay > 0) {
							Thread.sleep(TxDelay);
						}
/* JD Fix for TXing using C++ Modem
						if (Processor.TxModem.toString().startsWith("PSK")) {
							//Initialise PSK modems
							TXMpsk = new TxPSK(Processor.TxModem);
						} else if (Processor.TxModem.toString().startsWith("MFSK")) {
							//Initialise MFSK modems
							TXMmfsk = new TxMFSK(Processor.TxModem);
						} else if (Processor.TxModem.toString().startsWith("THOR")) {
							//Initialise THOR modems
							TXMthor = new TxTHOR(Processor.TxModem);
						} else {
							loggingclass.writelog("Wrong TX Mode called: " + Processor.TxModem.toString() , null, true);
						}

						//Send TX RSID if required
						if (Processor.TXID) {
							//Reset the send RSID flag
				    		Processor.TXID = false; 
							TxRSID.send();
						}

						if (Processor.TxModem.toString().startsWith("PSK")) {
							TXMpsk.AddBytes(TxSendline.getBytes());
						} else if (Processor.TxModem.toString().startsWith("MFSK")) {
							TXMmfsk.AddBytes(TxSendline.getBytes());
						} else if (Processor.TxModem.toString().startsWith("THOR")) {
							TXMthor.AddBytes(TxSendline.getBytes());
						} else {
							loggingclass.writelog("Wrong TX Mode called: " + Processor.TxModem.toString() , null, true);
						}
*/

						//Restart modem reception
						unPauseRxModem();
						Processor.TXActive = false;
						AndFlmsg.mHandler.post(AndFlmsg.updatetitle);   
					}
					catch (Exception e) {
						loggingclass.writelog("Can't output sound. Is Sound device busy?", null, true);
					}

				}

			}
		}
	};



	//Send Tune in a separate thread so that the UI thread is not blocked 
	//  during TX
	public static void TxTune() {
		Runnable TxTuneRun = new TxTuneThread();
		new Thread(TxTuneRun).start();
	}

	private static class TxTuneThread implements Runnable {
		private AudioTrack at = null; 

		public TxTuneThread() {
		}

		public void run() {
			//Stop the modem receiving side
			pauseRxModem();

			//Wait 1/2 second so that if there is potential RF feedback 
			//  on the touchscreen we do not start TXing while the 
			//  finger is still on the screen
			try {
				Thread.sleep(500);
			} catch (InterruptedException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			Processor.TXActive = true;
			String frequencySTR = config.getPreference("AFREQUENCY","1000");
			int frequency = Integer.parseInt(frequencySTR);

			int volumebits = Integer.parseInt(config.getPreference("VOLUME","8"));

			//Note the multiplier value for the buffer size
			int intSize = 4 * android.media.AudioTrack.getMinBufferSize(8000,
					AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT);

			if (AndFlmsg.toBluetooth) {
				//JD Bluetooth hack test
				//		        	at = new AudioTrack(AudioManager.STREAM_MUSIC, 8000, AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT, intSize , AudioTrack.MODE_STREAM);
				at = new AudioTrack(AudioManager.STREAM_VOICE_CALL, 8000, AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT, intSize , AudioTrack.MODE_STREAM);
			} else {
				at = new AudioTrack(AudioManager.STREAM_MUSIC, 8000, AudioFormat.CHANNEL_CONFIGURATION_MONO, AudioFormat.ENCODING_PCM_16BIT, intSize , AudioTrack.MODE_STREAM);
			}

			//Open audiotrack
			at.setStereoVolume(1.0f,1.0f);
			at.play();

			int sr = 8000; // should be active_modem->get_samplerate();
			int symlen = (int) (1 *  sr); //1 second buffer
			short[] outbuf = new short[symlen];

			double phaseincr;
			double phase = 0.0;
			phaseincr = 2.0 * Math.PI * frequency / sr;

			for (int i = 0; i < 3; i++) { //3 seconds tune
				for (int j = 0; j < symlen; j++) {
					phase += phaseincr;
					if (phase > 2.0 * Math.PI) phase -= 2.0 * Math.PI;
					outbuf[j] = (short) ((int) (Math.sin(phase) * 8386560) >> volumebits);
				}
				at.write(outbuf, 0, symlen);
			}

			//Stop audio track
			at.stop();
			//Wait for end of audio play to avoid 
			//overlaps between end of TX and start of RX
			while (at.getPlayState() == AudioTrack.PLAYSTATE_PLAYING) {
				try {
					Thread.sleep(50);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
				}
			}
			//Close audio track
			at.release();							
			Processor.TXActive = false;
			//Restart the modem receiving side
			unPauseRxModem();
		}
	};
}
